import { BrowserWindow, Rectangle, screen } from 'electron';
import { wm_create_conf, wm_move_conf, wm_resize_conf } from './../types/wm';
import { toP } from './utils';

interface WindowData {
  win: BrowserWindow;
  conf: wm_create_conf;
  follows?: string;
  followStuff?: any;
}

let wdata: { [key: string]: WindowData } = {};

const wm = {
  create: function (c: wm_create_conf) {
    const wid = Array.from({ length: 8 }, () => '0123456789abcdef'[Math.floor(Math.random() * 16)]).join('');
    const { width: sw, height: sh } = screen.getPrimaryDisplay().bounds;
    console.log('[wm] create', wid);
    const win = new BrowserWindow({
      width: c.whp ? toP(sw, c.w!) : c.w,
      height: c.whp ? toP(sh, c.h!) : c.h,
      x: c.xyp ? toP(sw, c.x!) : c.x,
      y: c.xyp ? toP(sh, c.y!) : c.y,
      frame: !c.noBorder,
      transparent: c.noBackground,
      webPreferences: {
        nodeIntegration: true,
        contextIsolation: false,
      },
    });

    wdata[wid] = { win: win, conf: c };
    win.setMenuBarVisibility(false); // tell me WHY should i NOT make this the default

    if (typeof c.onMinimize === 'function') win.on('minimize', () => c.onMinimize!(win));
    if (typeof c.onMaximize === 'function') win.on('maximize', () => c.onMaximize!(win));
    if (typeof c.onRestore === 'function') win.on('restore', () => c.onRestore!(win));
    if (typeof c.onFocus === 'function') win.on('focus', () => c.onFocus!(win));
    if (typeof c.onUnfocus === 'function') win.on('blur', () => c.onUnfocus!(win));
    if (typeof c.onClose === 'function') win.on('close', () => c.onClose!(win));

    if (c.onCreate) c.onCreate(win);
    return wid // just so you can reference it later on
  },
  destroy: async function (id: string) {
    console.log('[wm] destroy', id);
    const win = wdata[id];
    if (typeof win.conf.onDestroy === 'function') await win.conf.onDestroy!(win.win);
    win.win.destroy();
    delete wdata[id];
  },
  move: function (c: wm_move_conf) {
    console.log('[wm] move', c.id);
    const win = wdata[c.id];
    const { width: sw, height: sh } = screen.getPrimaryDisplay().bounds;
    let wb = win.win.getBounds();
  
    const sX = wb.x;
    const sY = wb.y;
    c.x = c.x === undefined ? sX : c.x;
    c.y = c.y === undefined ? sY : c.y;
  
    const duration = c.duration || 500;
    const tick = c.tick || 16;
    const totalSteps = Math.floor(duration / tick);

    let targetX = toP(c.p ? sw : 1, c.x);
    let targetY = toP(c.p ? sh : 1, c.y);
  
    if (c.fromCenter) {
      wb = win.win.getBounds();
      targetX = targetX - wb.width * 0.5;
      targetY = targetY - wb.height * 0.5;
    }
  
    const animate = () => {
      let currentStep = 0;
      const startX = sX;
      const startY = sY;
  
      const step = () => {
        currentStep++;
        const t = currentStep / totalSteps;
        const eT = c.ease ? c.ease(t) : t;
  
        const newX = startX + (targetX - startX) * eT;
        const newY = startY + (targetY - startY) * eT;
  
        win.win.setPosition(Math.round(newX), Math.round(newY));
  
        if (currentStep < totalSteps) {
          setTimeout(step, tick);
        } else {
          win.win.setPosition(Math.round(targetX), Math.round(targetY));
        }
      };
  
      step();
    };
  
    if (c.smooth) {
      animate();
    } else {
      win.win.setPosition(Math.round(targetX), Math.round(targetY));
    }
  },
  resize: function (c: wm_resize_conf) {
    console.log('[wm] resize', c.id);
    const win = wdata[c.id];
    if (!win || !win.win) {
      console.error('Window data not found for id:', c.id);
      return;
    }
    const wb = win.win.getBounds();
    const sW = wb.width, sH = wb.height;
    // Use the given dimensions if provided; otherwise fall back to the current window size
    c.w = c.w === undefined ? sW : c.w;
    c.h = c.h === undefined ? sH : c.h;
    const duration = c.duration || 500;
    const tick = c.tick || 16;
    const totalSteps = Math.floor(duration / tick);
    // Get primary display bounds – make sure that screen is imported from Electron
    const { width: sw, height: sh } = screen.getPrimaryDisplay().bounds;
    // Calculate target width and height using percentage conversion if needed.
    // Assuming toP converts a percentage value to pixels
    const targetW = toP(c.p ? sw : 1, c.w);
    const targetH = toP(c.p ? sh : 1, c.h);
  
    // Compute window center for centering if needed
    const centerX = wb.x + wb.width / 2;
    const centerY = wb.y + wb.height / 2;
  
    // Adjust position based on the chosen anchor.
    const getAnchoredPosition = (curWidth: number, curHeight: number) => {
      let newX = wb.x;
      let newY = wb.y;
      switch (c.anchor) {
        case 'top':
          // Align to top – x remains same
          newY = wb.y;
          break;
        case 'bottom':
          // Align to bottom
          newY = wb.y + wb.height - curHeight;
          break;
        case 'left':
          // Align to left – y remains same
          newX = wb.x;
          break;
        case 'right':
          // Align to right
          newX = wb.x + wb.width - curWidth;
          break;
        default:
          // Default: center the resized window relative to the original bounds
          newX = centerX - curWidth / 2;
          newY = centerY - curHeight / 2;
      }
      return { newX, newY };
    };
  
    // Animation function to smooth the resize
    const animate = () => {
      let currentStep = 0;
      const startW = wb.width;
      const startH = wb.height;
      const step = () => {
        currentStep++;
        const t = Math.min(currentStep / totalSteps, 1);
        const eT = c.ease ? c.ease(t) : t;
        const newW = startW + (targetW - startW) * eT;
        const newH = startH + (targetH - startH) * eT;
        // Get new positions based on the current width and height
        const pos = getAnchoredPosition(Math.round(newW), Math.round(newH));
        win.win.setBounds({
          x: pos.newX,
          y: pos.newY,
          width: Math.round(newW),
          height: Math.round(newH)
        });
        if (currentStep < totalSteps) {
          setTimeout(step, tick);
        } else {
          // Make sure to set final bounds exactly
          const finalPos = getAnchoredPosition(Math.round(targetW), Math.round(targetH));
          win.win.setBounds({
            x: finalPos.newX,
            y: finalPos.newY,
            width: Math.round(targetW),
            height: Math.round(targetH)
          });
        }
      };
      step();
    };
  
    // Use animate if smooth animation is requested, otherwise set bounds immediately.
    if (c.smooth) {
      animate();
    } else {
      const finalPos = getAnchoredPosition(Math.round(targetW), Math.round(targetH));
      win.win.setBounds({
        x: finalPos.newX,
        y: finalPos.newY,
        width: Math.round(targetW),
        height: Math.round(targetH)
      });
    }
  },  
  eval: async function (id: string, code: string) {
    console.log('[wm] eval', id);
    try {
      return await wdata[id].win.webContents.executeJavaScript(code);
    } catch (error) {
      console.error('Error executing JavaScript:', error);
    }
  },
  follow: {
    INTERNAL_follow: function (id: string) {
      const leadWin = wdata[wdata[id].follows!].win;
      const followWin = wdata[id].win;
      const leadBounds = leadWin.getBounds();
      const followBounds = followWin.getBounds();

      if (wdata[id].followStuff === undefined) {
        wdata[id].followStuff = {};
      }

      if (wdata[id].followStuff.offsetX === undefined || wdata[id].followStuff.offsetY === undefined) {
        wdata[id].followStuff.offsetX = followBounds.x - leadBounds.x;
        wdata[id].followStuff.offsetY = followBounds.y - leadBounds.y;
      }

      const newX = leadBounds.x + wdata[id].followStuff.offsetX;
      const newY = leadBounds.y + wdata[id].followStuff.offsetY;

      followWin.setPosition(newX, newY);
      followWin.show();
    },
    INTERNAL_followfocus: function (id: string) {
      wdata[id].win.show();
    },
    start: function (id: string, follows: string) {
      console.log('[wm]', id, 'follows', follows);
      wdata[id].follows = follows;
      wdata[follows].win.on('move', this.INTERNAL_follow.bind(this, id));
      wdata[follows].win.on('focus', this.INTERNAL_followfocus.bind(this, id));
    },    
    stop: function (id: string) {
      wdata[wdata[id].follows!].win.removeListener('move', this.INTERNAL_follow.bind(this, id));
      wdata[wdata[id].follows!].win.removeListener('focus', this.INTERNAL_followfocus.bind(this, id));
    }
  }
};

export default wm;
export { wdata, wm };